###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from auto_gen_objects import Object, Property, Choice

XML_TYPE = 'xml'

class XMLObject(Object):
    @staticmethod
    def convert(obj):
        if XML_TYPE in obj.layouts:
            obj.__class__ = XMLObject
            for prop in obj.properties:
                prop.__class__ = XMLProperty
            for choice in obj.choices:
                choice.__class__ = XMLChoice
                for prop in choice.properties:
                    prop.__class__ = XMLProperty

    def getName(self):
        try:
            return self.layouts[XML_TYPE]['name']
        except KeyError:
            pass
        return Object.getName(self)

    def getNodeType(self):
        try:
            return self.layouts[XML_TYPE]['nodeType']
        except KeyError:
            pass
        return 'xs:element'

    def getXMLAttributes(self):
        return [p for p in self.properties if p.hasSpec() and \
                    not p.isInferred() and p.getNodeType() == 'xs:attribute']

    def getXMLElements(self):
        return [p for p in self.properties if p.hasSpec() and \
                    not p.isInferred() and p.getNodeType() == 'xs:element']

    def getXMLChoices(self):
        return [c for c in self.choices if not c.isInverse() and \
                    any([p.hasSpec() for p in c.properties])]

    def isXMLChoice(self):
        return False

    def getXMLInferredProperties(self):
        return [p for p in self.properties if p.isInferred()]

    def getConstructorPairs(self):
        return [(f.getRegularName(), f.getRegularName()) 
                for f in self.getPythonFields() if f.hasSpec()]

class XMLProperty(Property):
    def hasSpec(self):
        return self.specs.has_key(XML_TYPE)

    def getName(self):
        try:
            return self.specs[XML_TYPE]['name']
        except KeyError:
            pass
        return Property.getName(self)

    def getNodeType(self):
        try:
            return self.specs[XML_TYPE]['nodeType']
        except KeyError:
            pass
        return 'xs:attribute'

    def getAttributeType(self):
        try:
            return self.specs[XML_TYPE]['type']
        except KeyError:
            pass
        return 'xs:string'

    def getAttributeUse(self):
        try:
            return self.specs[XML_TYPE]['use']
        except KeyError:
            pass
        return None

    def getAttributeUseText(self):
        if self.getAttributeUse() is not None:
            return ' use=%s' & self.getAttributeUse()
        return ''

    def getChoice(self):
        try:
            return self.specs[XML_TYPE]['choice']
        except KeyError:
            pass
        return None

    def isInferred(self):
        try:
            return self.specs[XML_TYPE]['inferred'] == 'true'
        except KeyError:
            pass
        return False

    def getXMLPropertyName(self):
        if self.isReference():
            refObj = self.getReferencedObject()
            return refObj.getName()
        return self.getName()

    def getMinOccurs(self):
        return '0'

    def getMaxOccurs(self):
        if self.isReference() and self.getMapping() != 'one-to-many':
            return '1'
        return 'unbounded'

class XMLChoice(Choice):
    def hasSpec(self):
        return self.properties[0].hasSpec()

    def getXMLProperties(self):
        return [p for p in self.properties if p.hasSpec()]

    def getXMLAttributes(self):
        return [p for p in self.properties if p.hasSpec() and \
                    not p.isInferred() and p.getNodeType() == 'xs:attribute']

    def getXMLElements(self):
        return [p for p in self.properties if p.hasSpec() and \
                    not p.isInferred() and p.getNodeType() == 'xs:element']

    def getXMLInferredProperties(self):
        return [p for p in self.properties if p.isInferred()]

    def isXMLChoice(self):
        return True

def convert(objects):
    xml_objects = []
    for obj in objects:
        if XML_TYPE in obj.layouts:
            XMLObject.convert(obj)
            xml_objects.append(obj)
    return xml_objects

def convert_schema_order(objects, root_type):
    ref_objects = []
    cur_objects = []
    for obj in objects:
        if obj.getName() == root_type:
            ref_objects = [obj]
            cur_objects = [obj]
            break
    if len(ref_objects) < 1:
        raise ValueError("Cannot find root %s" % root_type)
    while len(cur_objects) > 0:
        next_objects = []
        for obj in cur_objects:
            for prop in obj.getXMLElements():
                if prop.isReference() and \
                        prop.getReferencedObject() not in ref_objects and \
                        prop.getReferencedObject() not in next_objects:
                    next_objects.append(prop.getReferencedObject())
            for choice in obj.getXMLChoices():
                for prop in choice.getXMLElements():
                    if prop.isReference() and \
                            prop.getReferencedObject() not in ref_objects and \
                            prop.getReferencedObject() not in next_objects:
                        next_objects.append(prop.getReferencedObject())
        ref_objects.extend(next_objects)
        cur_objects = next_objects
    return ref_objects



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from PyQt4 import QtCore, QtGui

from vistrails.core.system import get_vistrails_basic_pkg_id
from vistrails.gui.modules.module_configure import StandardModuleConfigurationWidget

class ListConfigurationWidget(StandardModuleConfigurationWidget):
    """
    Configuration widget allowing to choose the number of ports.

    This is used to build a List from several modules while ensuring a given
    order. If no particular ordering is needed, connecting multiple ports to
    the 'head' input ports should be sufficient.

    """
    def __init__(self, module, controller, parent=None):
        """ ListConfigurationWidget(module: Module,
                                     controller: VistrailController,
                                     parent: QWidget)
                                     -> TupleConfigurationWidget

        Let StandardModuleConfigurationWidget constructor store the
        controller/module object from the builder and set up the
        configuration widget.
        After StandardModuleConfigurationWidget constructor, all of
        these will be available:
        self.module : the Module object int the pipeline
        self.controller: the current vistrail controller

        """
        StandardModuleConfigurationWidget.__init__(self, module,
                                                   controller, parent)

        # Give it a nice window title
        self.setWindowTitle("List Configuration")

        # Add an empty vertical layout
        centralLayout = QtGui.QVBoxLayout()
        centralLayout.setMargin(0)
        centralLayout.setSpacing(0)
        self.setLayout(centralLayout)

        # Add the configuration widget
        config_layout = QtGui.QFormLayout()
        self.number = QtGui.QSpinBox()
        self.number.setValue(self.countAdditionalPorts())
        self.connect(self.number, QtCore.SIGNAL('valueChanged(int)'),
                     lambda r: self.updateState())
        config_layout.addRow("Number of additional connections:", self.number)
        centralLayout.addLayout(config_layout)

        self.createButtons()

    def activate(self):
        self.number.focusWidget(QtCore.Qt.ActiveWindowFocusReason)

    def createButtons(self):
        """ createButtons() -> None
        Create and connect signals to Ok & Cancel button

        """
        self.buttonLayout = QtGui.QHBoxLayout()
        self.buttonLayout.setMargin(5)
        self.saveButton = QtGui.QPushButton("&Save", self)
        self.saveButton.setFixedWidth(100)
        self.saveButton.setEnabled(False)
        self.buttonLayout.addWidget(self.saveButton)
        self.resetButton = QtGui.QPushButton("&Reset", self)
        self.resetButton.setFixedWidth(100)
        self.resetButton.setEnabled(False)
        self.buttonLayout.addWidget(self.resetButton)
        self.layout().addLayout(self.buttonLayout)
        self.connect(self.saveButton, QtCore.SIGNAL('clicked(bool)'),
                     self.saveTriggered)
        self.connect(self.resetButton, QtCore.SIGNAL('clicked(bool)'),
                     self.resetTriggered)

    def saveTriggered(self, checked = False):
        """ saveTriggered(checked: bool) -> None
        Update vistrail controller and module when the user click Ok

        """
        if self.updateVistrail():
            self.saveButton.setEnabled(False)
            self.resetButton.setEnabled(False)
            self.state_changed = False
            self.emit(QtCore.SIGNAL('stateChanged'))
            self.emit(QtCore.SIGNAL('doneConfigure'), self.module.id)

    def closeEvent(self, event):
        self.askToSaveChanges()
        event.accept()

    def updateVistrail(self):
        """ updateVistrail() -> None
        Update Vistrail to contain changes in the port table

        """
        requested = self.number.value()
        current = self.countAdditionalPorts()

        if requested == current:
            # Nothing changed
            return

        if requested > current:
            sigstring = '(%s:Variant)' % get_vistrails_basic_pkg_id()
            add_ports = [('input', 'item%d' % i, sigstring, -1)
                           for i in xrange(current, requested)]
            self.controller.update_ports(self.module.id, [], add_ports)
        elif requested < current:
            delete_ports = [('input', p.name) for p in self.module.input_port_specs[requested-current:]]
            self.controller.update_ports(self.module.id, delete_ports, [])
        return True

    def countAdditionalPorts(self):
        return len(self.module.input_port_specs)

    def resetTriggered(self, checked = False):
        self.number.setValue(self.countAdditionalPorts())
        self.saveButton.setEnabled(False)
        self.resetButton.setEnabled(False)
        self.state_changed = False
        self.emit(QtCore.SIGNAL('stateChanged'))

    def updateState(self):
        if not self.hasFocus():
            self.setFocus(QtCore.Qt.TabFocusReason)
        self.saveButton.setEnabled(True)
        self.resetButton.setEnabled(True)
        if not self.state_changed:
            self.state_changed = True
            self.emit(QtCore.SIGNAL('stateChanged'))



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import os
import stat
from time import localtime
from datetime import datetime
from vistrails.core.thumbnails import ThumbnailCache

from entity import Entity

class ThumbnailEntity(Entity):
    type_id = 4

    def __init__(self, thumbnail=None):
        Entity.__init__(self)
        self.id = None
        self.update(thumbnail)

    @staticmethod
    def create(*args):
        entity = ThumbnailEntity()
        entity.load(*args)
        return entity

    def update(self, thumbnail):
        self.thumbnail = thumbnail
        if self.thumbnail is not None:
            # store in cache if not already there
            cache = ThumbnailCache.getInstance()
            cache._copy_thumbnails([thumbnail])
            self.name = os.path.basename(thumbnail)
            statinfo = os.stat(self.thumbnail)
            self.user = statinfo[stat.ST_UID]
            self.size = statinfo[stat.ST_SIZE]
            time = datetime(*localtime(statinfo[stat.ST_MTIME])[:6])
            self.mod_time = time
            self.create_time = time
            self.description = ""
            self.url = 'test'
            self.was_updated = True



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.modules.vistrails_module import Module
from vistrails.packages.spreadsheet.basic_widgets import SpreadsheetCell
from vistrails.packages.spreadsheet.spreadsheet_cell import QCellWidget

import qgis.core
import qgis.gui

import itertools
import os

class RasterLayer(Module):
    _input_ports = [('file', '(basic:File)'), 
                    ('name', '(basic:String)')]
    _output_ports = [('self', '(RasterLayer)')]

    def __init__(self):
        Module.__init__(self)
        self.qgis_obj = None

    def compute(self):
        fname = self.get_input('file').name
        if self.has_input('name'):
            name = self.get_input('name')
        else:
            name = os.path.splitext(os.path.basename(fname))[0]
        self.qgis_obj = qgis.core.QgsRasterLayer(fname, name)
        self.set_output('self', self)

class VectorLayer(Module):
    _input_ports = [('file', '(basic:File)'), 
                    ('name', '(basic:String)')]
    _output_ports = [('self', '(VectorLayer)')]

    def __init__(self):
        Module.__init__(self)
        self.qgis_obj = None

    def compute(self):
        fname = self.get_input('file').name
        if self.has_input('name'):
            name = self.get_input('name')
        else:
            name = os.path.splitext(os.path.basename(fname))[0]
        self.qgis_obj = qgis.core.QgsVectorLayer(fname, name, "ogr")
        self.set_output('self', self)

class QGISCell(SpreadsheetCell):
    """
    QGISCell is a VisTrails Module that can display QGIS inside a cell
    
    """
    _input_ports = [('rasterLayers', '(RasterLayer)'),
                    ('vectorLayers', '(VectorLayer)')]

    def __init__(self):
        SpreadsheetCell.__init__(self)
    
    def compute(self):
        rasterLayers = self.force_get_input_list('rasterLayers')
        vectorLayers = self.force_get_input_list('vectorLayers')
        self.displayAndWait(QGISCellWidget, (rasterLayers, vectorLayers))

class QGISCellWidget(QCellWidget):
    """
    QGISCellWidget is the actual QWidget taking the FigureManager
    as a child for displaying figures
    
    """
    def __init__(self, parent=None):
        """ QGISCellWidget(parent: QWidget) -> QGISCellWidget
        Initialize the widget with its central layout
        
        """
        QCellWidget.__init__(self, parent)
        centralLayout = QtGui.QVBoxLayout()
        self.setLayout(centralLayout)
        centralLayout.setMargin(0)
        centralLayout.setSpacing(0)

    def updateContents(self, inputPorts):
        """ updateContents(inputPorts: tuple) -> None
        Update the widget contents based on the input data
        
        """
        (rasterLayers, vectorLayers) = inputPorts

        canvas = qgis.gui.QgsMapCanvas()
        canvas.show()

        canvas.setCanvasColor(QtCore.Qt.white)
        canvas.enableAntiAliasing(True)

        for layer in itertools.chain(rasterLayers, vectorLayers):
            print 'adding layer', layer.qgis_obj
            qgis.core.QgsMapLayerRegistry.instance().addMapLayer(layer.qgis_obj)
        
        canvas.setExtent(rasterLayers[0].qgis_obj.extent())
        canvas.setLayerSet([qgis.gui.QgsMapCanvasLayer(layer.qgis_obj) 
                            for layer in itertools.chain(vectorLayers, 
                                                         rasterLayers)])

        # raster_fname = "/vistrails/local_packages/qgis/qgis_sample_data/raster/landcover.img"
        # raster = qgis.core.QgsRasterLayer(raster_fname, os.path.splitext(os.path.basename(raster_fname))[0])
        # qgis.core.QgsMapLayerRegistry.instance().addMapLayer(raster)

        # layer = qgis.core.QgsVectorLayer("/vistrails/local_packages/qgis/qgis_sample_data/vmap0_shapefiles/majrivers.shp", "Rivers", "ogr")
        # print layer.isValid()
        # qgis.core.QgsMapLayerRegistry.instance().addMapLayer(layer)
        # canvas.setLayerSet([qgis.gui.QgsMapCanvasLayer(layer), 
        #                     qgis.gui.QgsMapCanvasLayer(raster)])
    
        self.layout().addWidget(canvas)

        # Update the new figure canvas
        self.update()

    def grabWindowPixmap(self):
        """ grabWindowPixmap() -> QPixmap
        Widget special grabbing function
        
        """
        # return QtGui.QPixmap.grabWidget(self.figManager.canvas)
        return None

    def dumpToFile(self, filename):
        pass

    def saveToPDF(self, filename):
        pass

_modules = [RasterLayer, VectorLayer, QGISCell,]

def initialize():
    qgis.core.QgsApplication.setPrefixPath(
        "/vistrails/local_packages/qgis/QGIS.app/Contents/MacOS", True)
    qgis.core.QgsApplication.initQgis()




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" The file describes the parameter tree view

QParameterView
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from vistrails.core.inspector import PipelineInspector
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.gui.common_widgets import QSearchTreeWindow, QSearchTreeWidget
from vistrails.gui.paramexplore.pe_pipeline import QAnnotatedPipelineView
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface
from vistrails.core.utils import InstanceObject
from vistrails.core.debug import debug

################################################################################

class ParameterInfo(InstanceObject):
    #     ParameterInfo(type=,
    #                   identifier=,
    #                   namespace=,
    #                   value=,
    #                   id=,
    #                   dbtype=,
    #                   parent_dbtype=,
    #                   parent_id=,
    #                   is_alias=)
    #new: ParameterInfo(module_id=,
    #                   name=,
    #                   pos=,
    #                   value=,
    #                   spec=,
    #                   is_alias=)
    pass

################################################################################

class QParameterView(QtGui.QWidget, QVistrailsPaletteInterface):
    """
    QParameterView contains the parameter exploration properties and the
    parameter palette
    
    """
    def __init__(self, controller=None, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.set_title('Pipeline Methods')
        
        self.controller = None
        vLayout = QtGui.QVBoxLayout()
        vLayout.setMargin(0)
        vLayout.setSpacing(5)
        self.setLayout(vLayout)

        self.toggleUnsetParameters = QtGui.QCheckBox('Show Unset Parameters')
        vLayout.addWidget(self.toggleUnsetParameters, 0, QtCore.Qt.AlignRight)

        self.parameterWidget = QParameterWidget()
        vLayout.addWidget(self.parameterWidget)
        self.treeWidget = self.parameterWidget.treeWidget

        self.pipeline_view = QAnnotatedPipelineView()
        self.pipeline_view.setReadOnlyMode(True)
        vLayout.addWidget(self.pipeline_view)

        vLayout.setStretch(0,0)
        vLayout.setStretch(1,1)
        vLayout.setStretch(2,0)

        self.connect(self.toggleUnsetParameters, QtCore.SIGNAL("toggled(bool)"),
                     self.parameterWidget.treeWidget.toggleUnsetParameters)
        self.set_controller(controller)

    def set_controller(self, controller):
        if self.controller == controller:
            return
        self.controller = controller
        self.pipeline_view.set_controller(controller)
        if self.controller is not None:
            self.set_pipeline(controller.current_pipeline)
        else:
            self.set_pipeline(None)

    def set_pipeline(self, pipeline):
        if self.controller is None:
            return
        self.pipeline = pipeline
        self.parameterWidget.set_pipeline(pipeline, self.controller)
        if pipeline:
            self.pipeline_view.scene().setupScene(pipeline)
        else:
            self.pipeline_view.scene().clear()
        self.pipeline_view.updateAnnotatedIds(pipeline)
        # this ensures the correct pipeline is set when updating exploration
        self.get_palette().set_pipeline(pipeline)

    def get_palette(self):
        from vistrails.gui.paramexplore.pe_inspector import QParamExploreInspector
        return QParamExploreInspector.instance()

class QParameterWidget(QSearchTreeWindow):
    """
    QParameterWidget is a special widget for displaying aliases and
    parameters inside a pipeline
    
    """
    def createTreeWidget(self):
        """ createTreeWidget() -> QModuleTreeWidget
        Return the search tree widget for this window
        
        """
        treeWidget = QParameterTreeWidget(self)
        return treeWidget

    def set_pipeline(self, pipeline, controller):
        self.pipeline = pipeline
        self.treeWidget.updateFromPipeline(pipeline, controller)


class QParameterTreeWidget(QSearchTreeWidget):
    """
    QParameterTreeWidget is a subclass of QSearchTreeWidget to display all
    Vistrails Module
    
    """
    def __init__(self, parent=None):
        """ QParameterTreeWidget(parent: QWidget) -> QParameterTreeWidget
        Set up size policy and header

        """
        QSearchTreeWidget.__init__(self, parent)
        self.header().hide()
        self.setRootIsDecorated(False)
        self.delegate = QParameterTreeWidgetItemDelegate(self, self)
        self.setItemDelegate(self.delegate)
        self.showUnsetParameters = False

    def updateFromPipeline(self, pipeline, controller):
        """ updateFromPipeline(pipeline: Pipeline) -> None
        Read the list of aliases and parameters from the pipeline
        
        """
        self.clear()
        if not pipeline:
            return

        # Update the aliases
        if len(pipeline.aliases)>0:
            aliasRoot = QParameterTreeWidgetItem(None, self, ['Aliases'])
            aliasRoot.setFlags(QtCore.Qt.ItemIsEnabled)
            for (alias, info) in pipeline.aliases.iteritems():
                ptype, pId, parentType, parentId, mId = info
                parameter = pipeline.db_get_object(ptype, pId)
                function = pipeline.db_get_object(parentType, parentId)
                v = parameter.strValue
                port_spec = function.get_spec('input')
                port_spec_item = port_spec.port_spec_items[parameter.pos]
                label = ['%s = %s' % (alias, v)]
                pInfo = ParameterInfo(module_id=mId,
                                      name=function.name,
                                      pos=parameter.pos,
                                      value=v,
                                      spec=port_spec_item,
                                      is_alias=True)
                QParameterTreeWidgetItem((alias, [pInfo]),
                                         aliasRoot, label)
            aliasRoot.setExpanded(True)

        vistrailVarsRoot = QParameterTreeWidgetItem(None, self,
                                      ['Vistrail Variables'])
        vistrailVarsRoot.setHidden(True)

        # Now go through all modules and functions

        inspector = PipelineInspector()
        inspector.inspect_ambiguous_modules(pipeline)
        sortedModules = sorted(pipeline.modules.iteritems(),
                               key=lambda item: item[1].name)

        reg = get_module_registry()

        for mId, module in sortedModules:
            if module.is_vistrail_var():
                vistrailVarsRoot.setHidden(False)
                vistrailVarsRoot.setExpanded(True)
                port_spec = module.get_port_spec('value', 'input')
                if not port_spec:
                    debug.critical("Not port_spec for value in module %s" % module)
                    continue

                if not controller.has_vistrail_variable_with_uuid(
                                        module.get_vistrail_var()):
                    continue
                vv = controller.get_vistrail_variable_by_uuid(
                                        module.get_vistrail_var())


                label = ['%s = %s' % (vv.name, vv.value)]
                pList = [ParameterInfo(module_id=mId,
                                       name=port_spec.name,
                                       pos=port_spec.port_spec_items[pId].pos,
                                       value="",
                                       spec=port_spec.port_spec_items[pId],
                                       is_alias=False)
                         for pId in xrange(len(port_spec.port_spec_items))]
                QParameterTreeWidgetItem((vv.name, pList),
                                         vistrailVarsRoot,
                                         label)
                continue
                
            function_names = {}
            # Add existing parameters
            mLabel = [module.name]
            moduleItem = None
            if len(module.functions)>0:
                for fId in xrange(len(module.functions)):
                    function = module.functions[fId]
                    function_names[function.name] = function
                    if len(function.params)==0: continue
                    if moduleItem==None:
                        if inspector.annotated_modules.has_key(mId):
                            annotatedId = inspector.annotated_modules[mId]
                            moduleItem = QParameterTreeWidgetItem(annotatedId,
                                                                  self, mLabel)
                        else:
                            moduleItem = QParameterTreeWidgetItem(None,
                                                                  self, mLabel)
                    v = ', '.join([p.strValue for p in function.params])
                    label = ['%s(%s)' % (function.name, v)]
                    
                    try:
                        port_spec = function.get_spec('input')
                    except Exception:
                        debug.critical("get_spec failed: %s %s %s" % \
                                       (module, function, function.sigstring))
                        continue
                    port_spec_items = port_spec.port_spec_items
                    pList = [ParameterInfo(module_id=mId,
                                           name=function.name,
                                           pos=function.params[pId].pos,
                                           value=function.params[pId].strValue,
                                           spec=port_spec_items[pId],
                                           is_alias=False)
                             for pId in xrange(len(function.params))]
                    mName = module.name
                    if moduleItem.parameter is not None:
                        mName += '(%d)' % moduleItem.parameter
                    fName = '%s :: %s' % (mName, function.name)
                    QParameterTreeWidgetItem((fName, pList),
                                             moduleItem,
                                             label)
            # Add available parameters
            if module.is_valid:
                for port_spec in module.destinationPorts():
                    if (port_spec.name in function_names or
                        not port_spec.is_valid or 
                        not len(port_spec.port_spec_items) or
                        not reg.is_constant(port_spec)):
                        # The function already exists or is empty
                        # or contains non-constant modules
                        continue
                    if moduleItem==None:
                        if inspector.annotated_modules.has_key(mId):
                            annotatedId = inspector.annotated_modules[mId]
                            moduleItem = QParameterTreeWidgetItem(annotatedId,
                                                                  self, 
                                                                  mLabel, 
                                                                  False)
                        else:
                            moduleItem = QParameterTreeWidgetItem(None, self,
                                                                  mLabel, False)
                    v = ', '.join([p.module for p in port_spec.port_spec_items])
                    label = ['%s(%s)' % (port_spec.name, v)]
                    pList = [ParameterInfo(module_id=mId,
                                           name=port_spec.name,
                                           pos=port_spec.port_spec_items[pId].pos,
                                           value="",
                                           spec=port_spec.port_spec_items[pId],
                                           is_alias=False)
                             for pId in xrange(len(port_spec.port_spec_items))]
                    mName = module.name
                    if moduleItem.parameter is not None:
                        mName += '(%d)' % moduleItem.parameter
                    fName = '%s :: %s' % (mName, port_spec.name)
                    QParameterTreeWidgetItem((fName, pList),
                                             moduleItem,
                                             label, False)
            if moduleItem:
                moduleItem.setExpanded(True)
        self.toggleUnsetParameters(self.showUnsetParameters)

    def toggleUnsetParameters(self, state):
        self.showUnsetParameters = state
        for item in self.findItems("*", QtCore.Qt.MatchWildcard | QtCore.Qt.MatchRecursive):
            if not item.isSet:
                item.setHidden(not state)
            
class QParameterTreeWidgetItemDelegate(QtGui.QItemDelegate):
    """    
    QParameterTreeWidgetItemDelegate will override the original
    QTreeWidget paint function to draw buttons for top-level item
    similar to QtDesigner. This mimics
    Qt/tools/designer/src/lib/shared/sheet_delegate, which is only a
    private class from QtDesigned.
    
    """
    def __init__(self, view, parent):
        """ QParameterTreeWidgetItemDelegate(view: QTreeView,
                                          parent: QWidget)
                                          -> QParameterTreeWidgetItemDelegate
        Create the item delegate given the tree view
        
        """
        QtGui.QItemDelegate.__init__(self, parent)
        self.treeView = view

    def paint(self, painter, option, index):
        """ painter(painter: QPainter, option QStyleOptionViewItem,
                    index: QModelIndex) -> None
        Repaint the top-level item to have a button-look style
        
        """
        model = index.model()
        if not model.parent(index).isValid():
            style = self.treeView.style()
            r = option.rect
            textrect = QtCore.QRect(r.left() + 10,
                                    r.top(),
                                    r.width() - 10,
                                    r.height())
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
            text = option.fontMetrics.elidedText(
                model.data(index, QtCore.Qt.DisplayRole),
                QtCore.Qt.ElideMiddle, 
                textrect.width()-10)
            style.drawItemText(painter,
                               textrect,
                               QtCore.Qt.AlignLeft,
                               option.palette,
                               self.treeView.isEnabled(),
                               text)
            painter.setPen(QtGui.QPen(QtCore.Qt.black))
            fm = QtGui.QFontMetrics(font)
            size = fm.size(QtCore.Qt.TextSingleLine, text)
            painter.drawLine(textrect.left()-5,
                             textrect.bottom()-1,
                             textrect.left()+size.width()+5,
                             textrect.bottom()-1)

            annotatedId = model.data(index, QtCore.Qt.UserRole+1)            
            if annotatedId:
                idRect = QtCore.QRect(
                    QtCore.QPoint(textrect.left()+size.width()+5,
                                  textrect.top()),
                    textrect.bottomRight())
                QAnnotatedPipelineView.drawId(painter, idRect,
                                              annotatedId,
                                              QtCore.Qt.AlignLeft |
                                              QtCore.Qt.AlignVCenter)
        else:
            QtGui.QItemDelegate.paint(self, painter, option, index)

    def sizeHint(self, option, index):
        """ sizeHint(option: QStyleOptionViewItem, index: QModelIndex) -> None
        Take into account the size of the top-level button
        
        """
        return (QtGui.QItemDelegate.sizeHint(self, option, index) +
                QtCore.QSize(2, 2))
            

class QParameterTreeWidgetItem(QtGui.QTreeWidgetItem):
    """
    QParameterTreeWidgetItem represents module on QParameterTreeWidget
    
    """
    def __init__(self, info, parent, labelList, isSet=True):
        """ QParameterTreeWidgetItem(info: (str, []),
                                     parent: QTreeWidgetItem
                                     labelList: string,
                                     isSet: bool)
                                     -> QParameterTreeWidget
                                     
        Create a new tree widget item with a specific parent and
        labels. info describing a set of paramters as follow:
        (name, [ParameterInfo]):
           name  = Name of the parameter set (alias or function)
        If this item is a top-level item, info can either be None or
        an integer specifying the annotated id of this module
        isSet indicates if it represents a set or unset parameter
        """
        self.parameter = info
        QtGui.QTreeWidgetItem.__init__(self, parent, labelList)
        if isinstance(self.parameter, int):
            self.setData(0, QtCore.Qt.UserRole+1,
                         self.parameter)
        self.isSet = isSet




#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This will build the html and the pdf version of the user's guide and put
them in where defined. 
Please configure the variables below according to your server's setup

"""
import os
import os.path
import re
import shutil
import subprocess
import sys
import time
### Begin configuration ###

# Should we build the pdf version
BUILD_PDF = True

# Should we build the html version
BUILD_HTML = False

# Folder where vistrails is
PATH_TO_VISTRAILS_GIT = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))

# Folder where the html files will be placed. This script will build
# the docs in the standard location and move files later
HTML_FOLDER = None

# Complete file path to where copy final pdf file
PDF_FILE = os.path.join(PATH_TO_VISTRAILS_GIT, "scripts", "VisTrails.pdf")

# Should we run a `git pull` before building docs? 
PERFORM_GIT_PULL = False

### End configuration ### 
### The following variables usually don't need to be changed
USERSGUIDE_SUBPATH = ['doc', 'usersguide']
GIT_PULL_CMD = ["git", "pull"]
BUILD_HTML_SUBPATH = ["_build", "html"]
BUILD_LATEX_SUBPATH = ["_build", "latex"]
PDF_BUILD_NAME = "VisTrails.pdf"
if len(sys.argv)>1:
    PDF_FILE = os.path.abspath(sys.argv[1])
    if PDF_FILE[-4:] != '.pdf':
        PDF_FILE = os.path.join(PDF_FILE, PDF_BUILD_NAME)
if __name__ == '__main__':
    if (PATH_TO_VISTRAILS_GIT is not None and
        os.path.exists(PATH_TO_VISTRAILS_GIT)):
        current_folder = os.getcwd()
        if PERFORM_GIT_PULL:
            print "performing git pull in %s..."%PATH_TO_VISTRAILS_GIT
            os.chdir(PATH_TO_VISTRAILS_GIT)
            proc = subprocess.Popen(GIT_PULL_CMD,
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT)
            proc.wait()
            if proc.returncode != 0:
                print "ERROR: git pull failed."
                if proc.stdout:
                    print proc.stdout.readlines()
        
        # The api needs path to source set
        env = os.environ.copy()
        env['PYTHONPATH'] = PATH_TO_VISTRAILS_GIT

        os.chdir(os.path.join(PATH_TO_VISTRAILS_GIT,
                              *USERSGUIDE_SUBPATH))
        
        print "Going into directory %s"%os.getcwd()
        if BUILD_HTML:
            # now build html documentation    
            print "now building html documentation..."
            proc = subprocess.Popen(["make", "html"],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)
            proc.wait()
            if proc.returncode != 0:
                print "ERROR: make html failed."
                if proc.stdout:
                    print proc.stdout.readlines()
            else:
                # now move files to their final place
                if HTML_FOLDER is not None:
                    if os.path.exists(HTML_FOLDER):
                        shutil.rmtree(HTML_FOLDER)
                    html_build = os.path.join(os.getcwd(),
                                              *BUILD_HTML_SUBPATH)
                    shutil.move(html_build, HTML_FOLDER)

        if BUILD_PDF:
            print "Building usersguide to ", PDF_FILE
            #build latex files
            print "now preparing latex files..."
            latex_path = os.path.join(".",
                                  *BUILD_LATEX_SUBPATH)
            if os.path.exists(latex_path):
                print "removing old %s directory..."%latex_path
                proc = subprocess.Popen(["rm", "-rf", latex_path],
                                        stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT)
                proc.wait()
            print "building latex files..."
            proc = subprocess.Popen(["make", "latex"],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)
            proc.wait()
            if proc.returncode != 0:
                print "ERROR: make latex failed."
                if proc.stdout:
                    print proc.stdout.readlines()
            else:
                #now build pdf
                print "now building pdf file..."
                os.chdir(os.path.join(os.getcwd(),
                         *BUILD_LATEX_SUBPATH))
                proc = subprocess.Popen(["make", "LATEXOPTS=-interaction=nonstopmode -halt-on-error", "all-pdf"],
                                    stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)
                output, error = proc.communicate()
                if proc.returncode != 0:
                    print "ERROR: make all-pdf failed."
                    print output, error
                else:
                    #now move file to its final place
                    if PDF_FILE is not None:
                        if os.path.exists(PDF_FILE):
                            os.unlink(PDF_FILE)
                        pdf_build = os.path.join(os.getcwd(),
                                                 PDF_BUILD_NAME)
                        shutil.move(pdf_build, PDF_FILE)
        os.chdir(current_folder)
        print "Done."
    else:
        print "PATH_TO_VISTRAILS_GIT was not provided. Exiting."



#!/usr/bin/env python

# Create a constrained Delaunay triangulation following fault lines. The
# fault lines serve as constraint edges in the Delaunay triangulation.

import vtk
from vtk.util.misc import vtkGetDataRoot
from vtk.util.colors import *
VTK_DATA_ROOT = vtkGetDataRoot()

# Generate some points by reading a VTK data file. The data file also
# has edges that represent constraint lines. This is originally from a
# geologic horizon.
reader = vtk.vtkPolyDataReader()
reader.SetFileName(VTK_DATA_ROOT + "/Data/faults.vtk")

# Perform a 2D triangulation with constraint edges.
delny = vtk.vtkDelaunay2D()
delny.SetInputConnection(reader.GetOutputPort())
delny.SetSourceConnection(reader.GetOutputPort())
delny.SetTolerance(0.00001)
normals = vtk.vtkPolyDataNormals()
normals.SetInputConnection(delny.GetOutputPort())
mapMesh = vtk.vtkPolyDataMapper()
mapMesh.SetInputConnection(normals.GetOutputPort())
meshActor = vtk.vtkActor()
meshActor.SetMapper(mapMesh)
meshActor.GetProperty().SetColor(beige)

# Now pretty up the mesh with tubed edges and balls at the vertices.
tuber = vtk.vtkTubeFilter()
tuber.SetInputConnection(reader.GetOutputPort())
tuber.SetRadius(25)
mapLines = vtk.vtkPolyDataMapper()
mapLines.SetInputConnection(tuber.GetOutputPort())
linesActor = vtk.vtkActor()
linesActor.SetMapper(mapLines)
linesActor.GetProperty().SetColor(1, 0, 0)
linesActor.GetProperty().SetColor(tomato)

# Create graphics objects
# Create the rendering window, renderer, and interactive renderer
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(linesActor)
ren.AddActor(meshActor)
ren.SetBackground(1, 1, 1)
renWin.SetSize(350, 250)

cam1 = vtk.vtkCamera()
cam1.SetClippingRange(2580, 129041)
cam1.SetFocalPoint(461550, 6.58e+006, 2132)
cam1.SetPosition(463960, 6.559e+06, 16982)
cam1.SetViewUp(-0.321899, 0.522244, 0.78971)
light = vtk.vtkLight()
light.SetPosition(0, 0, 1)
light.SetFocalPoint(0, 0, 0)
ren.SetActiveCamera(cam1)
ren.AddLight(light)

ren.GetActiveCamera().Zoom(1.5)
iren.LightFollowCameraOff()

iren.Initialize()
renWin.Render()
iren.Start()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.modules.vistrails_module import Module, InvalidOutput, \
    ModuleError
import copy

#################################################################################
## If Operator

class If(Module):
    """
    The If Module alows the user to choose the part of the workflow to be
    executed through the use of a condition.
    """

    def update_upstream(self):
        """A modified version of the update_upstream method."""

        # everything is the same except that we don't update anything
        # upstream of TruePort or FalsePort
        excluded_ports = set(['TruePort', 'FalsePort', 'TrueOutputPorts',
                              'FalseOutputPorts'])
        for port_name, connector_list in self.inputPorts.iteritems():
            if port_name not in excluded_ports:
                for connector in connector_list:
                    connector.obj.update()
        for port_name, connectorList in copy.copy(self.inputPorts.items()):
            if port_name not in excluded_ports:
                for connector in connectorList:
                    if connector.obj.get_output(connector.port) is \
                            InvalidOutput:
                        self.remove_input_connector(port_name, connector)

    def compute(self):
        """ The compute method for the If module."""

        if not self.has_input('Condition'):
            raise ModuleError(self, 'Must set condition')
        cond = self.get_input('Condition')

        if cond:
            port_name = 'TruePort'
            output_ports_name = 'TrueOutputPorts'
        else:
            port_name = 'FalsePort'
            output_ports_name = 'FalseOutputPorts'

        if self.has_input(output_ports_name):
            for connector in self.inputPorts.get(output_ports_name):
                connector.obj.update()

        if not self.has_input(port_name):
            raise ModuleError(self, 'Must set ' + port_name)

        for connector in self.inputPorts.get(port_name):
            connector.obj.update()

            if self.has_input(output_ports_name):
                output_ports = self.get_input(output_ports_name)
                result = []
                for output_port in output_ports:
                    result.append(connector.obj.get_output(output_port))

                # FIXME can we just make this a list?
                if len(output_ports) == 1:
                    self.set_output('Result', result[0])
                else:
                    self.set_output('Result', result)

#################################################################################
## Default module

class Default(Module):
    """
    The Default module allows the user to provide a default value.

    This module can be put in the middle of a connection to provide a default
    value from the Default port in case nothing is set on the Input port. This
    is particularly useful when using subworkflows, with InputPort modules with
    optional set to True.
    """

    def compute(self):
        if self.has_input('Input'):
            self.set_output('Result', self.get_input('Input'))
        else:
            self.set_output('Result', self.get_input('Default'))


###############################################################################

import unittest
import urllib2

from vistrails.tests.utils import intercept_result, execute

class TestIf(unittest.TestCase):
    def do_if(self, val):
        with intercept_result(If, 'Result') as results:
            interp_dict = execute([
                    ('If', 'org.vistrails.vistrails.control_flow', [
                        ('FalseOutputPorts', [('List', "['value']")]),
                        ('TrueOutputPorts', [('List', "['value']")]),
                        ('Condition', [('Boolean', str(val))]),
                    ]),
                    ('Integer', 'org.vistrails.vistrails.basic', [
                        ('value', [('Integer', '42')]),
                    ]),
                    ('Integer', 'org.vistrails.vistrails.basic', [
                        ('value', [('Integer', '28')]),
                    ]),
                ],
                [
                    (1, 'self', 0, 'TruePort'),
                    (2, 'self', 0, 'FalsePort'),
                ],
                full_results=True)
            self.assertFalse(interp_dict.errors)
        if val:
            self.assertEqual(results, [42])
        else:
            self.assertEqual(results, [28])
        self.assertEqual(interp_dict.executed, {0: True, 1: val, 2: not val})

    def test_if_true(self):
        self.do_if(True)

    def test_if_false(self):
        self.do_if(False)


class TestDefault(unittest.TestCase):
    def do_default(self, val):
        if val:
            src = 'o = 42'
        else:
            src = ('from vistrails.core.modules.vistrails_module import '
                   'InvalidOutput\n'
                   'o = InvalidOutput')
        src = urllib2.quote(src)
        with intercept_result(Default, 'Result') as results:
            self.assertFalse(execute([
                    ('Default', 'org.vistrails.vistrails.control_flow', [
                        ('Default', [('Integer', '28')]),
                    ]),
                    ('PythonSource', 'org.vistrails.vistrails.basic', [
                        ('source', [('String', src)]),
                    ]),
                ],
                [
                    (1, 'o', 0, 'Input'),
                ],
                add_port_specs=[
                    (1, 'output', 'o',
                     'org.vistrails.vistrails.basic:Integer'),
                ]))
        if val:
            self.assertEqual(results, [42])
        else:
            self.assertEqual(results, [28])

    def test_default_set(self):
        self.do_default(True)

    def test_default_unset(self):
        self.do_default(False)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

import os
import shutil
import subprocess

try:
    from ctypes import windll, Structure, c_ulong, c_ulonglong, byref, sizeof
    importSuccess = True

    class WIN32MEMORYSTATUSEX(Structure):
        """ Structure that represents memory information returned by
        Windows API

        """
        _fields_ = [
            ('dwLength', c_ulong),
            ('dwMemoryLoad', c_ulong),
            ('dwTotalPhys', c_ulonglong),
            ('dwAvailPhys', c_ulonglong),
            ('dwTotalPageFile', c_ulonglong),
            ('dwAvailPageFile', c_ulonglong),
            ('dwTotalVirtual', c_ulonglong),
            ('dwAvailVirtual', c_ulonglong),
            ('dwAvailExtendedVirtual', c_ulonglong),
            ]

except ImportError:
    importSuccess = False


__all__ = ['guess_total_memory', 'home_directory',
           'list2cmdline', 'remote_copy_program', 'remote_shell_program',
           'graph_viz_dot_command_line', 'remove_graph_viz_temporaries',
           'link_or_copy', 'executable_is_in_path', 'execute_cmdline',
           'get_executable_path', 'execute_piped_cmdlines', 'execute_cmdline2',
           'shell_font_face', 'shell_font_size',
           'TestWindows']


##############################################################################
def parse_meminfo():
    """
    parse_meminfo() -> long
    Calls Windows 32 API GlobalMemoryStatus(Ex) to get memory information
    It requires ctypes module

    """
    try:
        kernel32 = windll.kernel32

        result = WIN32MEMORYSTATUSEX()
        result.dwLength = sizeof(WIN32MEMORYSTATUSEX)
        kernel32.GlobalMemoryStatusEx(byref(result))
    except Exception:
        return -1
    return result.dwTotalPhys // 1024

def guess_total_memory():
    """ guess_total_memory() -> int
    Return system memory in megabytes. If ctypes is not installed it returns -1

    """
    if importSuccess:
        return parse_meminfo()
    else:
        return -1

def home_directory():
    """ home_directory() -> str
    Returns user's home directory using windows environment variables
    $HOMEDRIVE and $HOMEPATH

    """
    if len(os.environ['HOMEPATH']) == 0:
        return '\\'
    else:
        return os.environ['HOMEDRIVE'] + os.environ['HOMEPATH']

def remote_copy_program():
    return "pscp -P"

def remote_shell_program():
    return "plink -P"

def graph_viz_dot_command_line():
    """ graph_viz_dot_command_line() -> str
    Returns dot command line

    """
    return 'dot -Tplain -o'

def remove_graph_viz_temporaries():
    pass

def link_or_copy(src, dst):
    """link_or_copy(src:str, dst:str) -> None
    Copies file src to dst

    """
    shutil.copyfile(src, dst)

def executable_is_in_path(filename):
    """ executable_is_in_path(filename: str) -> bool
    Check if exename can be reached in the PATH environment.
    """
    pathlist = os.environ['PATH'].split(os.pathsep) + ["."]
    exts = os.environ['PATHEXT'].split(os.pathsep)
    for path in pathlist:
        for ext in exts:
            fullpath = os.path.join(path, filename) + ext
            if os.path.isfile(fullpath):
                return True
    return False

def list2cmdline(lst):
    for el in lst:
        assert isinstance(el, basestring)
    return subprocess.list2cmdline(lst)

def execute_cmdline(lst, output):
    """execute_cmdline(lst: list of str)-> int Builds a command line
    enquoting the arguments properly and executes it using popen4. It
    returns the output on output. popen4 doesn't return a code, so it
    will always return 0

    """
    proc = subprocess.Popen(lst, shell=True, stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    proc.wait()
    if proc.stdout:
        output.extend(proc.stdout.readlines())
    return proc.returncode

def get_executable_path(executable_name):
    """get_executable_path(executable_name: str) -> str
    Get the absolute filename of an executable, searching in VisTrails's top
    directory then the PATH.
    """
    exts = os.environ['PATHEXT'].split(os.pathsep)

    # Search in top directory
    filename = os.path.abspath(os.path.join(
            os.path.dirname(__file__),
            '../../..',
            executable_name))
    for ext in exts:
        if os.path.isfile(filename + ext):
            return filename

    # Search in path
    pathlist = os.environ['PATH'].split(os.pathsep) + ["."]
    for path in pathlist:
        for ext in exts:
            fullpath = os.path.join(path, executable_name) + ext
            if os.path.isfile(fullpath):
                return os.path.abspath(fullpath)

def execute_piped_cmdlines(cmd_list_list):
    stdin = subprocess.PIPE
    for cmd_list in cmd_list_list:
        cmd_line = list2cmdline(cmd_list)
        process = subprocess.Popen(cmd_line, shell=True,
                                   stdin=stdin,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        stdin = process.stdout
    (output, errs) = process.communicate()
    result = process.returncode
    return (result, output, errs)

def execute_cmdline2(cmd_list):
    return execute_piped_cmdlines([cmd_list])

def shell_font_face():
    return 'Courier New'

def shell_font_size():
    return 8

################################################################################

import unittest

class TestWindows(unittest.TestCase):
    """ Class to test Windows specific functions """

    def test1(self):
        """ Test if guess_total_memory() is returning an int >= 0"""
        result = guess_total_memory()
        assert isinstance(result, (int, long))
        assert result >= 0

    def test2(self):
        """ Test if home_directory is not empty """
        result = home_directory()
        assert result != ""

    def test_executable_file_in_path(self):
        self.assertTrue(executable_is_in_path('cmd'))

if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""Routines common to Linux and OSX."""
from __future__ import division

import os
import subprocess


__all__ = ['executable_is_in_path', 'list2cmdline', 'execute_cmdline',
           'get_executable_path', 'execute_piped_cmdlines', 'execute_cmdline2']

###############################################################################

def executable_is_in_path(filename):
    """ executable_is_in_path(filename: str) -> string
    Check if exename can be reached in the PATH environment.
    """
    pathlist = os.environ['PATH'].split(os.pathsep) + ["."]
    for path in pathlist:
        fullpath = os.path.join(path, filename)
        if os.path.isfile(fullpath):
            return True
    return False

def list2cmdline(lst):
    for el in lst:
        assert isinstance(el, basestring)
    return subprocess.list2cmdline(lst)

def execute_cmdline(lst, output):
    """execute_cmdline(lst: list of str, output)-> int
    Builds a command line enquoting the arguments properly and executes it
    using subprocess.Popen. It returns the error code and the output is on 'output'.

    cscheid: why don't we return a tuple of int, lst instead of mutating that list?

    """
    process = subprocess.Popen(lst, shell=False,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               close_fds=True)
    result = process.wait()
    output.extend(process.stdout.readlines())
    return result

def get_executable_path(executable_name):
    """get_executable_path(executable_name: str) -> str
    Get the absolute filename of an executable, searching in the PATH.
    """
    pathlist = os.environ['PATH'].split(os.pathsep)
    for path in pathlist:
        fullpath = os.path.join(path, executable_name)
        if os.path.isfile(fullpath):
            return os.path.abspath(fullpath)

def execute_piped_cmdlines(cmd_list_list):
    stdin = subprocess.PIPE
    for cmd_list in cmd_list_list:
        cmd_line = list2cmdline(cmd_list)
        process = subprocess.Popen(cmd_line, shell=True,
                                   stdin=stdin,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   close_fds=True)
        stdin = process.stdout
    (output, errs) = process.communicate()
    result = process.returncode
    return (result, output, errs)

def execute_cmdline2(cmd_list):
    return execute_piped_cmdlines([cmd_list])



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This module contains class definitions for:

    * ModuleFunction
"""
from __future__ import division

from vistrails.db.domain import DBFunction
from vistrails.core.modules.utils import create_port_spec_string
from vistrails.core.utils import enum, VistrailsInternalError, all, eprint
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.core.vistrail.port_spec import PortSpec
from itertools import izip
import copy

import unittest
import copy
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.db.domain import IdScope
import vistrails.core

################################################################################

class ModuleFunction(DBFunction):
    __fields__ = ['name', 'returnType', 'params']
    """ Stores a function from a vistrail module """

    ##########################################################################
    # Constructors and copy
    
    def __init__(self, *args, **kwargs):
        DBFunction.__init__(self, *args, **kwargs)
        if self.name is None:
            self.name = ""
        if self.real_id is None:
            self.real_id = -1
        if self.pos is None:
            self.pos = -1
        self.set_defaults()
        
    def set_defaults(self, other=None):
        if other is None:
            self.returnType = "void"
            self.is_valid = False
        else:
            self.returnType = other.returnType
            self.is_valid = other.is_valid
        self.parameter_idx = self.db_parameters_id_index

    def __copy__(self):
        """ __copy__() -> ModuleFunction - Returns a clone of itself """
        return ModuleFunction.do_copy(self)

    def do_copy(self, new_ids=False, id_scope=None, id_remap=None):
        cp = DBFunction.do_copy(self, new_ids, id_scope, id_remap)
        cp.__class__ = ModuleFunction
        cp.set_defaults(self)
        return cp

    @staticmethod
    def convert(_function):
        if _function.__class__ == ModuleFunction:
            return
        _function.__class__ = ModuleFunction
        for _parameter in _function.db_get_parameters():
            ModuleParam.convert(_parameter)
        _function.set_defaults()

    ##########################################################################
    # Properties

    id = DBFunction.db_pos
    pos = DBFunction.db_pos
    real_id = DBFunction.db_id
    name = DBFunction.db_name   

    def _get_sigstring(self):
        return create_port_spec_string([p.spec_tuple for p in self.params])
    sigstring = property(_get_sigstring)

    def _get_params(self):
        self.db_parameters.sort(key=lambda x: x.db_pos)
        return self.db_parameters
    def _set_params(self, params):
        self.db_parameters = params
    # If you're mutating the params property, watch out for the sort
    # gotcha: every time you use the params on reading position,
    # they get resorted
    params = property(_get_params, _set_params)
    parameters = property(_get_params, _set_params)

    def add_parameter(self, param):
        self.db_add_parameter(param)
    addParameter = add_parameter

    def add_parameters(self, params):
        for p in params:
            self.db_add_parameter(p)

    ##########################################################################

    def getNumParams(self):
        """ getNumParams() -> int Returns the number of params. """
        return len(self.params)
    
    def serialize(self, doc, element):
        """serialize(doc, element) -> None - Writes itself in XML """
        child = doc.createElement('function')
        child.setAttribute('name',self.name)
        child.setAttribute('returnType',self.type)
        for p in self.params:
            p.serialize(doc,child)
        element.appendChild(child)

    def get_spec(self, port_type):
        """ get_spec(port_type) -> PortSpec

        Returns a PortSpec corresponding to the function parameter
        types set.  This is useful to make module functions look more
        like they are 'regular' modules and connections (which is what
        they get compiled down to in execution).

        port_type is either 'input' or 'output', as strings, which
        simply gets set on the spec being returned.
        """
        assert port_type == 'input' or port_type == 'output'
        result = PortSpec(sigstring=self.sigstring)
        result.type = port_type
        return result

    ##########################################################################
    # Debugging

    def show_comparison(self, other):
        if type(self) != type(other):
            print "type mismatch"
            return
        if self.name != other.name:
            print "name mismatch"
            return
        if self.returnType != other.returnType:
            print "return type mismatch"
            return
        if len(self.params) != len(other.params):
            print "params length mismatch"
            return
        for p,q in izip(self.params, other.params):
            if p != q:
                print "params mismatch"
                p.show_comparison(q)
                return
        print "no difference found"
        assert self == other
        return

    ##########################################################################
    # Operators
    
    def __str__(self):
        """ __str__() -> str - Returns a string representation of itself """
        return ("<function id='%s' pos='%s' name='%s' params=%s)@%X" %
                (self.real_id,
                 self.pos,
                 self.name,
                 [str(p) for p in self.params],
                 id(self)))

    def __eq__(self, other):
        """ __eq__(other: ModuleFunction) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if type(self) != type(other):
            return False
        if self.name != other.name:
            return False
        if self.returnType != other.returnType:
            return False
        if len(self.params) != len(other.params):
            return False
        for p,q in zip(self.params, other.params):
            if p != q:
                return False
        return True
            
    def __ne__(self, other):
        """ __ne__(other: ModuleFunction) -> boolean
        Returns True if self and other don't have the same attributes. 
        Used by !=  operator. 
        
        """
        return not self.__eq__(other)

################################################################################
# Testing


#TODO add more meaningful tests

class TestModuleFunction(unittest.TestCase):

    def create_function(self, id_scope=IdScope()):
        param = ModuleParam(id=id_scope.getNewId(ModuleParam.vtType),
                            pos=2,
                            type='Int',
                            val='1')
        function = ModuleFunction(id=id_scope.getNewId(ModuleFunction.vtType),
                                  pos=0,
                                  name='value',
                                  parameters=[param])
        return function

    def test_copy(self):        
        id_scope = IdScope()
        f1 = self.create_function(id_scope)
        f2 = copy.copy(f1)
        self.assertEquals(f1, f2)
        self.assertEquals(f1.id, f2.id)
        f3 = f1.do_copy(True, id_scope, {})
        self.assertEquals(f1, f3)
        self.assertNotEquals(f1.real_id, f3.real_id)

    def test_serialization(self):
        import vistrails.core.db.io
        f1 = self.create_function()
        xml_str = vistrails.core.db.io.serialize(f1)
        f2 = vistrails.core.db.io.unserialize(xml_str, ModuleFunction)
        self.assertEquals(f1, f2)
        self.assertEquals(f1.real_id, f2.real_id)
                            
    def testComparisonOperators(self):
        f = ModuleFunction()
        f.name = "value"
        param = ModuleParam()
        param.strValue = "1.2"
        param.type = "Float"
        param.alias = ""
        f.addParameter(param)
        g = ModuleFunction()
        g.name = "value"
        param = ModuleParam()
        param.strValue = "1.2"
        param.type = "Float"
        param.alias = ""
        g.addParameter(param)
        assert f == g
        param = ModuleParam()
        param.strValue = "1.2"
        param.type = "Float"
        param.alias = ""
        g.addParameter(param)
        assert f != g

    def test_str(self):
        f = ModuleFunction(name='value',
                           parameters=[ModuleParam(type='Float',
                                                   val='1.2')],
                           )
        str(f)

if __name__ == '__main__':
    unittest.main()
    



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" RecentVistrailList is a Helper class to manage serialization and
unserialization of a list of locators to XML """
from __future__ import division

from vistrails.core.system import get_elementtree_library
from vistrails.core.db.locator import FileLocator, DBLocator

ElementTree = get_elementtree_library()

class RecentVistrailList(object):
    def __init__(self):
        self.maxlocators = 0
        self.locators = []
        self.locators_map = {}
        
    @staticmethod
    def unserialize(text):
        """ unserialize(text) -> RecentVistrailList """
        root = ElementTree.fromstring(text)
        if root.tag != 'recentVistrails':
            return None
        vtlist = RecentVistrailList()
        for node in root.getchildren():
            loc = FileLocator.from_xml(node)
            if loc is None:
                loc = DBLocator.from_xml(node, include_name=True)
            if loc is not None:
                vtlist.locators.append(loc)
                vtlist.locators_map[loc.name] = loc
        return vtlist
    
    def serialize(self, node=None):
        """serialize(node: ElementTree.Element) -> text
        Convert this object to an XML representation.
        """
        if node is None:
            node = ElementTree.Element('recentVistrails')
        for loc in self.locators:
            childNode = ElementTree.SubElement(node, 'locator')
            if isinstance(loc, DBLocator):
                loc.to_xml(childNode, include_name=True)
            else:
                loc.to_xml(childNode)
        return ElementTree.tostring(node)
    
    def ensure_no_more_than_max(self, max):
        while len(self.locators) > max:
            locator = self.locators.pop()
            del self.locators_map[locator.name]
            
        self.maxlocators = max
        
    def add_locator(self, locator):
        if locator in self.locators:
            self.locators.remove(locator)
        self.locators.insert(0, locator)
        self.locators_map[locator.name] = locator
        
        self.ensure_no_more_than_max(self.maxlocators)
        
    def remove_locator(self, locator):
        if locator in self.locators:
            self.locators.remove(locator)
            del self.locators_map[locator.name]
            
    def length(self):
        return len(self.locators)
    
    def get_locator(self, index):
        try:
            return self.locators[index]
        except IndexError:
            return None
        
    def get_locator_by_name(self, name):
        try:
            return self.locators_map[name]
        except KeyError:
            return None


###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file describe a widget for keeping version notes,info and tag
name

QVersionProp
QVersionNotes
QVersionPropOverlay
QExpandButton
QNotesDialog
QVersionThumbs
QVersionMashups

"""
from __future__ import division

import re
from PyQt4 import QtCore, QtGui
from vistrails.core.configuration import get_vistrails_configuration
from vistrails.core import debug
from vistrails.core.thumbnails import ThumbnailCache
from vistrails.core.utils import all
from vistrails.core.vistrail.controller import custom_color_key, \
    parse_custom_color
from vistrails.gui.theme import CurrentTheme
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

###############################################################################

class ColorChooserButton(QtGui.QPushButton):
    color_selected = QtCore.pyqtSignal(object)

    def __init__(self, parent=None):
        QtGui.QToolButton.__init__(self, parent)
        self.setColor(None)

        self.connect(self, QtCore.SIGNAL('clicked()'), self.changeColor)

    def setColor(self, color, silent=True):
        self.color = color
        if color is not None:
            self.setStyleSheet('ColorChooserButton {'
                               'border: 1px solid black; '
                               'background-color: rgb(%d, %d, %d); }' % (
                               color.red(), color.green(), color.blue()))
        else:
            self.setStyleSheet('ColorChooserButton {'
                               'border: 1px dashed black; }')
        self.update()
        if not silent:
            self.color_selected.emit(self.color)

    def sizeHint(self):
        return QtCore.QSize(20, 20)

    def changeColor(self):
        if self.color is not None:
            self.setColor(None, silent=False)
        else:
            color = QtGui.QColorDialog.getColor(QtCore.Qt.white, self)
            if color.isValid():
                self.setColor(color, silent=False)

###############################################################################

class QVersionProp(QtGui.QWidget, QVistrailsPaletteInterface):
    """
    QVersionProp is a widget holding property of a version including
    tagname and notes
    
    """    
    def __init__(self, parent=None):
        """ QVersionProp(parent: QWidget) -> QVersionProp
        Initialize the main layout
        
        """
        QtGui.QWidget.__init__(self, parent)
        self.setWindowTitle('Workflow Info')

        vLayout = QtGui.QVBoxLayout()
        vLayout.setMargin(2)
        vLayout.setSpacing(2)
        self.setLayout(vLayout)

        gLayout = QtGui.QGridLayout()
        gLayout.setMargin(2)
        gLayout.setSpacing(5)
        gLayout.setColumnMinimumWidth(1,5)
        gLayout.setRowMinimumHeight(0,20)
        gLayout.setRowMinimumHeight(1,20)
        gLayout.setRowMinimumHeight(2,20)
        gLayout.setRowMinimumHeight(3,20)
        vLayout.addLayout(gLayout)
        
        tagLabel = QtGui.QLabel('Tag:', self)
        gLayout.addWidget(tagLabel, 0, 0, 1, 1)

        editLayout = QtGui.QHBoxLayout()
        editLayout.setMargin(2)
        editLayout.setSpacing(2)
        self.tagEdit = QtGui.QLineEdit()
        tagLabel.setBuddy(self.tagEdit)
        editLayout.addWidget(self.tagEdit)
        self.tagEdit.setEnabled(False)
        self.tagEdit.setMinimumHeight(22)

        self.tagReset = QtGui.QToolButton(self)
        self.tagReset.setIcon(QtGui.QIcon(
                self.style().standardPixmap(QtGui.QStyle.SP_DialogCloseButton)))
        self.tagReset.setIconSize(QtCore.QSize(12,12))
        self.tagReset.setAutoRaise(True)
        self.tagReset.setEnabled(False)
        editLayout.addWidget(self.tagReset)

        configuration = get_vistrails_configuration()
        self.use_custom_colors = configuration.check('enableCustomVersionColors')

        if self.use_custom_colors:
            self.customColor = ColorChooserButton(self)
            editLayout.addWidget(self.customColor)
            self.customColor.color_selected.connect(self.custom_color_selected)

        gLayout.addLayout(editLayout, 0, 2, 1, 1)

        userLabel = QtGui.QLabel('User:', self)
        gLayout.addWidget(userLabel, 1, 0, 1, 1)
        
        self.userEdit = QtGui.QLabel('', self)
        gLayout.addWidget(self.userEdit, 1, 2, 1, 1)

        dateLabel = QtGui.QLabel('Date:', self)
        gLayout.addWidget(dateLabel, 2, 0, 1, 1)

        self.dateEdit = QtGui.QLabel('', self)
        gLayout.addWidget(self.dateEdit, 2, 2, 1, 1)

        idLabel = QtGui.QLabel('ID:', self)
        gLayout.addWidget(idLabel, 3, 0, 1, 1)
        
        self.idEdit = QtGui.QLabel('', self)
        gLayout.addWidget(self.idEdit, 3, 2, 1, 1)

        self.notesLabel = QtGui.QLabel('Notes:')
        gLayout.addWidget(self.notesLabel, 4, 0, 1, 1)

        self.versionNotes = QVersionNotes()
        vLayout.addWidget(self.versionNotes)
        self.versionNotes.setEnabled(False)

        self.versionThumbs = QVersionThumbs()
        vLayout.addWidget(self.versionThumbs)
        
        self.versionMashups = QVersionMashups()
        vLayout.addWidget(self.versionMashups)
                
        self.connect(self.tagEdit, QtCore.SIGNAL('editingFinished()'),
                     self.tagFinished)
        self.connect(self.tagEdit, QtCore.SIGNAL('textChanged(QString)'),
                     self.tagChanged)
        self.connect(self.tagReset, QtCore.SIGNAL('clicked()'),
                     self.tagCleared)

        self.controller = None
        self.versionNumber = -1

    def updateController(self, controller):
        """ updateController(controller: VistrailController) -> None
        Assign the controller to the property page
        
        """
        if self.controller == controller:
            return
        self.controller = controller
        self.versionNotes.controller = controller
        self.versionThumbs.controller = controller
        self.versionMashups.controller = controller
        if controller is not None:
            self.updateVersion(controller.current_version)
        else:
            self.updateVersion(-1)

    def updateVersion(self, versionNumber):
        """ updateVersion(versionNumber: int) -> None
        Update the property page of the version
        
        """
        self.versionNumber = versionNumber
        self.versionNotes.updateVersion(versionNumber)
        self.versionThumbs.updateVersion(versionNumber)
        self.versionMashups.updateVersion(versionNumber)
        if self.controller:
            vistrail = self.controller.vistrail
            if self.use_custom_colors:
                custom_color = vistrail.get_action_annotation(versionNumber,
                                                              custom_color_key)
                if custom_color is not None:
                    try:
                        custom_color = parse_custom_color(custom_color.value)
                        custom_color = QtGui.QColor(*custom_color)
                    except ValueError, e:
                        debug.warning("Version %r has invalid color "
                                      "annotation (%s)" % (versionNumber, e))
                        custom_color = None
                self.customColor.setColor(custom_color)

            if vistrail.actionMap.has_key(versionNumber):
                # Follow upgrades forward to find tag
                tag = vistrail.search_upgrade_versions(
                        versionNumber,
                        lambda vt, v, bv: vt.getVersionName(v) or None) or ''

                if getattr(get_vistrails_configuration(), 'hideUpgrades', True):
                    base_ver = vistrail.get_base_upgrade_version(versionNumber)
                else:
                    base_ver = versionNumber

                action = vistrail.actionMap[base_ver]
                self.tagEdit.setText(tag)
                self.userEdit.setText(action.user)
                self.dateEdit.setText(action.date)
                self.idEdit.setText(unicode(action.id))
                self.tagEdit.setEnabled(True)
                return
            else:
                self.tagEdit.setEnabled(False)
                self.tagReset.setEnabled(False)

        self.tagEdit.setText('')
        self.userEdit.setText('')
        self.dateEdit.setText('')
        self.idEdit.setText('')

    def tagFinished(self):
        """ tagFinished() -> None
        Update the new tag to vistrail
        
        """
        if self.controller:
            name = self.controller.vistrail.getVersionName(self.versionNumber)
            currentText = str(self.tagEdit.text())
            if name != currentText:    
                self.controller.update_current_tag(currentText)

    def tagChanged(self, text):
        """ tagChanged(text: QString) -> None
        Update the button state if there is text

        """
        self.tagReset.setEnabled(text != '')

    def tagCleared(self):
        """ tagCleared() -> None
        Remove the tag
        
        """ 
        self.tagEdit.setText('')
        self.tagFinished()

    def custom_color_selected(self, color):
        if color is not None:
            self.controller.vistrail.set_action_annotation(
                    self.controller.current_version, custom_color_key,
                    '%d,%d,%d' % (color.red(), color.green(), color.blue()))
        else:
            self.controller.vistrail.set_action_annotation(
                    self.controller.current_version, custom_color_key, None)
        self.controller.set_changed(True)
        self.controller.recompute_terse_graph()
        self.controller.invalidate_version_tree()

class QVersionNotes(QtGui.QTextEdit):
    """
    QVersionNotes is text widget that update/change a version notes
    
    """    
    def __init__(self, parent=None):
        """ QVersionNotes(parent: QWidget) -> QVersionNotes
        Initialize control variables
        
        """
        QtGui.QTextEdit.__init__(self, parent)
        self.controller = None
        self.versionNumber = -1
        self.update_on_focus_out = True
        self.setAcceptRichText(False)
        # Reset text to black, for some reason it is grey by default on the mac
        self.palette().setBrush(QtGui.QPalette.Text,
                                QtGui.QBrush(QtGui.QColor(0,0,0,255)))

    def updateVersion(self, versionNumber):
        """ updateVersion(versionNumber: int) -> None
        Update the text to be the notes of the vistrail versionNumber
        
        """
        if self.versionNumber == versionNumber:
            return
        self.versionNumber = versionNumber
        if self.controller:
            if self.controller.vistrail.actionMap.has_key(versionNumber):
                # Follow upgrades forward to find notes
                notes = self.controller.vistrail.search_upgrade_versions(
                        versionNumber,
                        lambda vt, v, bv: vt.get_notes(v) or None)
                if notes:
                    self.setHtml(notes)
                    # work around a strange bug where an empty new paragraph gets added every time
                    self.trim_first_paragraph()
                else:
                    self.setHtml('')
                self.setEnabled(True)
                return
            else:
                self.setEnabled(False)
        self.setPlainText('')

    def commit_changes(self):
        if self.controller and self.document().isModified():
            self.controller.update_notes('' if self.document().isEmpty()
                                         else str(self.toHtml()))

    def reset_changes(self):
        """ reset_changes() -> None

        """
        self.updateVersion(self.versionNumber)

    def focusOutEvent(self, event):
        """ focusOutEvent(event: QFocusEvent) -> None
        Update the version notes if the text has been modified
        
        """
        if self.update_on_focus_out:
            self.commit_changes()
        QtGui.QTextEdit.focusOutEvent(self,event)

    def trim_first_paragraph(self):
        doc = self.document()
        count = doc.blockCount()
        cursor = QtGui.QTextCursor(doc)
        cursor.select(QtGui.QTextCursor.LineUnderCursor)
        sel = cursor.selection().toPlainText()
        if all(c == ' ' for c in sel):
            cursor.removeSelectedText()
            cursor = QtGui.QTextCursor(doc)
            cursor.deleteChar()

###############################################################################

class QVersionPropOverlay(QtGui.QFrame):
    """
    QVersionPropOverlay is a transparent widget that sits on top of the version
    view.  It displays properties of a version: tag, user, date, and notes.

    """
    propagatingEvent = set([
        QtCore.QEvent.MouseButtonDblClick,
        QtCore.QEvent.MouseButtonPress,
        QtCore.QEvent.MouseButtonRelease,
        QtCore.QEvent.MouseMove,
        QtCore.QEvent.Wheel,
        ])
    def __init__(self, parent=None, mouseWidget=None):
        """ QVersionPropOverlay(parent: QWidget) -> QVersionPropOverlay
        Setup layout

        """
        QtGui.QFrame.__init__(self, parent)
        self.propagatingWidget = mouseWidget
        self.palette = QtGui.QPalette()
        self.palette.setColor(QtGui.QPalette.Base, QtGui.QColor(0,0,0,0))
        self.setPalette(self.palette)
        self.setAutoFillBackground(True)
        self.setFrameStyle(QtGui.QFrame.NoFrame)
        self.setFrameShadow(QtGui.QFrame.Plain)
        self.layout = QtGui.QGridLayout()
        self.layout.setVerticalSpacing(1)

        self.tag_label = QtGui.QLabel()
        self.tag_label.palette().setBrush(QtGui.QPalette.Text,
                                          CurrentTheme.VERSION_PROPERTIES_PEN)
        self.tag_label.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)
        self.tag_label.setText("Tag:")

        self.tag = QtGui.QLabel()
        self.tag.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)

        self.description_label = QtGui.QLabel()
        self.description_label.palette().setBrush(QtGui.QPalette.Text,
                                                  CurrentTheme.VERSION_PROPERTIES_PEN)
        self.description_label.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)
        self.description_label.setText("Action:")

        self.description = QtGui.QLabel()
        self.description.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)

        self.user_label = QtGui.QLabel()
        self.user_label.palette().setBrush(QtGui.QPalette.Text,
                                           CurrentTheme.VERSION_PROPERTIES_PEN)
        self.user_label.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)
        self.user_label.setText("User:")

        self.user = QtGui.QLabel()
        self.user.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)

        self.date_label = QtGui.QLabel()
        self.date_label.palette().setBrush(QtGui.QPalette.Text,
                                           CurrentTheme.VERSION_PROPERTIES_PEN)
        self.date_label.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)
        self.date_label.setText("Date:")
        
        self.date = QtGui.QLabel()
        self.date.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)

        self.notes_label = QtGui.QLabel()
        self.notes_label.palette().setBrush(QtGui.QPalette.Text,
                                           CurrentTheme.VERSION_PROPERTIES_PEN)
        self.notes_label.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)
        self.notes_label.setText("Notes:")

        self.notes = QtGui.QLabel()
        self.notes.setTextFormat(QtCore.Qt.PlainText)
        self.notes.setFont(CurrentTheme.VERSION_PROPERTIES_FONT)
        
        self.notes_button = QExpandButton()
        self.notes_button.hide()

        self.notes_layout = QtGui.QHBoxLayout()
        self.notes_layout.addWidget(self.notes)
        self.notes_layout.addWidget(self.notes_button)
        self.notes_layout.addStretch()

        self.layout.addWidget(self.tag_label, 0, 0)
        self.layout.addWidget(self.tag, 0, 1)
        self.layout.addWidget(self.description_label, 1, 0)
        self.layout.addWidget(self.description, 1, 1)
        self.layout.addWidget(self.user_label, 2, 0)
        self.layout.addWidget(self.user, 2, 1)
        self.layout.addWidget(self.date_label, 3, 0)
        self.layout.addWidget(self.date, 3, 1)
        self.layout.addWidget(self.notes_label, 4, 0)
        self.layout.addLayout(self.notes_layout, 4, 1)

        self.layout.setColumnMinimumWidth(0,35)
        self.layout.setColumnMinimumWidth(1,200)
        self.layout.setContentsMargins(2,2,2,2)
        self.layout.setColumnStretch(1,1)
        self.setLayout(self.layout)
        self.updateGeometry()
        self.controller = None
        
        self.notes_dialog = QNotesDialog(self)
        self.notes_dialog.hide()

        QtCore.QObject.connect(self.notes_button,
                               QtCore.SIGNAL("pressed()"),
                               self.openNotes)

    def updateGeometry(self):
        """ updateGeometry() -> None
        Keep in upper left of screen

        """
        parentGeometry = self.parent().geometry()
        self.pos_x = 4
        self.pos_y = 4
        self.move(self.pos_x, self.pos_y)

    def updateController(self, controller):
        """ updateController(controller: VistrailController) -> None
        Assign the controller to the properties
        
        """
        if self.controller == controller:
            return
        self.controller = controller
        self.notes_dialog.updateController(controller)

    def updateVersion(self, versionNumber):
        """ updateVersion(versionNumber: int) -> None
        Update the text items
        
        """

        self.notes_dialog.updateVersion(versionNumber)
        if self.controller:
            if self.controller.vistrail.actionMap.has_key(versionNumber):
                vistrail = self.controller.vistrail
                # Follow upgrades forward to find tag
                tag = vistrail.search_upgrade_versions(
                        versionNumber,
                        lambda vt, v, bv: vt.getVersionName(v) or None) or ''

                if getattr(get_vistrails_configuration(), 'hideUpgrades', True):
                    base_ver = vistrail.get_base_upgrade_version(versionNumber)
                else:
                    base_ver = versionNumber

                action = vistrail.actionMap[base_ver]
                description = vistrail.get_description(base_ver)
                self.tag.setText(self.truncate(tag))
                self.description.setText(self.truncate(description))
                self.user.setText(self.truncate(action.user))
                self.date.setText(self.truncate(action.date))
                notes = vistrail.get_notes(action.id)
                if notes:
                    s = self.convertHtmlToText(notes)
                    self.notes.setText(self.truncate(s))
                else:
                    self.notes.setText('')
                self.notes_button.show()
            else:
                self.tag.setText('')
                self.description.setText('')
                self.user.setText('')
                self.date.setText('')
                self.notes.setText('')
                self.notes_button.hide()

    def convertHtmlToText(self, str):
        """ convertHtmlToText(str: QString) -> QString
        Remove HTML tags and newlines
        
        """
        # Some text we want to ignore lives outside brackets in the header
        str = re.sub(r"<head>.*</head>", r"", str)
        # Remove all other tags
        str = re.sub(r"<[^>]*>", r"", str)
        # Remove newlines
        str = re.sub(r"\n", r"", str)
        return str

    def truncate(self, str):
        """ truncate(str: string) -> string
        Shorten string to fit in smaller space
        
        """
        if len(str) > 24:
            str = str[:22] + "..."
        return str

    def openNotes(self):
        """ openNotes() -> None

        Show notes dialog
        """
        self.notes_dialog.show()
        self.notes_dialog.activateWindow()

    def event(self, e):
        """ Propagate all mouse events to the right widget """
        if e.type() in self.propagatingEvent:
            if self.propagatingWidget!=None:
                QtCore.QCoreApplication.sendEvent(self.propagatingWidget, e)
                return False
        return QtGui.QFrame.event(self, e)

###############################################################################

class QExpandButton(QtGui.QLabel):
    """
    A transparent button type with a + draw in 
    """
    def __init__(self, parent=None):
        """
        QExpandButton(parent: QWidget) -> QExpandButton
        """
        QtGui.QLabel.__init__(self, parent)
        
        self.drawButton(0)
        self.setToolTip('Edit Notes')
        self.setScaledContents(False)
        self.setFrameShape(QtGui.QFrame.NoFrame)

    def sizeHint(self):
        """ sizeHint() -> QSize
        
        """
        return QtCore.QSize(10,10)
        

    def mousePressEvent(self, e):
        """ mousePressEvent(e: QMouseEvent) -> None
        Capture mouse press event on the frame to move the widget
        
        """
        if e.buttons() & QtCore.Qt.LeftButton:
            self.drawButton(1)
    
    def mouseReleaseEvent(self, e):
        self.drawButton(0)
        self.emit(QtCore.SIGNAL('pressed()'))

    def drawButton(self, down):
        """ drawButton(down: bool) -> None
        Custom draw function
        
        """
        self.picture = QtGui.QPicture()
        painter = QtGui.QPainter()
        painter.begin(self.picture)
        painter.setRenderHints(QtGui.QPainter.Antialiasing, False)
        pen = QtGui.QPen(QtCore.Qt.SolidLine)
        pen.setWidth(1)
        pen.setCapStyle(QtCore.Qt.RoundCap)
        brush = QtGui.QBrush(QtCore.Qt.NoBrush)
        if (down):
            pen.setColor(QtGui.QColor(150,150,150,150))
        else:
            pen.setColor(QtGui.QColor(0,0,0,255))

        painter.setPen(pen)
        painter.setBrush(brush)
        painter.drawRect(0,0,8,8)
        painter.drawLine(QtCore.QLine(4,2,4,6))
        painter.drawLine(QtCore.QLine(2,4,6,4))
        painter.end()
        self.setPicture(self.picture)

###############################################################################

class QNotesDialog(QtGui.QDialog):
    """
    A small non-modal dialog with text entry to modify and view notes

    """
    def __init__(self, parent=None):
        """
        QNotesDialog(parent: QWidget) -> QNotesDialog

        """
        QtGui.QDialog.__init__(self, parent)
        self.setModal(False)
        self.notes = QVersionNotes(self)
        self.notes.update_on_focus_out = False
        layout = QtGui.QVBoxLayout(self)
        layout.addWidget(self.notes)
        layout.setMargin(0)

        self.apply_button = QtGui.QPushButton('Apply', self)
        self.apply_button.setDefault(False)
        self.apply_button.setAutoDefault(False)
        self.ok_button = QtGui.QPushButton('Ok', self)
        self.ok_button.setDefault(False)
        self.ok_button.setAutoDefault(False)
        self.cancel_button = QtGui.QPushButton('Cancel', self)
        self.cancel_button.setDefault(False)
        self.cancel_button.setAutoDefault(False)
        self.buttonLayout = QtGui.QHBoxLayout()
        layout.addLayout(self.buttonLayout)
        self.buttonLayout.addWidget(self.apply_button)
        self.buttonLayout.addWidget(self.ok_button)
        self.buttonLayout.addWidget(self.cancel_button)

        self.setLayout(layout)
        self.controller = None

        QtCore.QObject.connect(self.apply_button,
                               QtCore.SIGNAL("released()"),
                               self.apply_pressed)

        QtCore.QObject.connect(self.ok_button,
                               QtCore.SIGNAL("released()"),
                               self.ok_pressed)

        QtCore.QObject.connect(self.cancel_button,
                               QtCore.SIGNAL("released()"),
                               self.cancel_pressed)
    def apply_pressed(self):
        """ apply_pressed() -> None

        """
        self.notes.commit_changes()

    def ok_pressed(self):
        """ ok_pressed() -> None
        
        """
        self.notes.commit_changes()
        self.close()

    def cancel_pressed(self):
        """ cancel_pressed() -> None
        
        """
        self.notes.reset_changes()
        self.close()

    def updateController(self, controller):
        """ updateController(controller: VistrailController) -> None

        """
        self.controller = controller
        self.notes.controller = controller

    def updateVersion(self, versionNumber):
        """ updateVersion(versionNumber: int) -> None

        """
        self.notes.updateVersion(versionNumber)
        if self.controller:
            if self.controller.vistrail.actionMap.has_key(versionNumber):
                # Follow upgrades forward to find tag
                tag = self.controller.vistrail.search_upgrade_versions(
                        versionNumber,
                        lambda vt, v, bv: vt.getVersionName(v) or None) or ''

                title = "Notes: " + tag
                self.setWindowTitle(title)
            else:
                self.setWindowTitle("Notes")

    def sizeHint(self):
        """ sizeHint() -> QSize
        
        """
        return QtCore.QSize(250,200)
        
###############################################################################

class QVersionThumbs(QtGui.QWidget):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.versionNumber = None
        label = QtGui.QLabel("Preview:")
        self.thumbs = QtGui.QLabel()
        self.thumbs.setFrameShape(QtGui.QFrame.NoFrame)
        self.scrollArea = QtGui.QScrollArea(self)
        self.scrollArea.setFrameStyle(QtGui.QFrame.NoFrame)
        self.scrollArea.setWidget(self.thumbs)
        self.scrollArea.setWidgetResizable(True)
        layout = QtGui.QVBoxLayout()
        layout.setMargin(2)
        layout.addWidget(label)
        layout.addWidget(self.scrollArea)
        self.setLayout(layout)
        self.controller = None
        
    def updateController(self, controller):
        """ updateController(controller: VistrailController) -> None

        """
        self.controller = controller
        
    def updateVersion(self, versionNumber):
        """ updateVersion(versionNumber: int) -> None

        """
        if self.controller:
            vistrail = self.controller.vistrail
            if versionNumber in vistrail.actionMap.keys():
                # Follow upgrades forward to find a thumbnail
                thumb_ver = self.controller.vistrail.search_upgrade_versions(
                        versionNumber,
                        lambda vt, v, bv: v if vt.has_thumbnail(v) else None)
                if thumb_ver is not None:
                    action = vistrail.actionMap[thumb_ver]
                    if vistrail.has_thumbnail(action.id):
                        cache = ThumbnailCache.getInstance()
                        fname = cache.get_abs_name_entry(
                            vistrail.get_thumbnail(action.id))
                        if fname is not None:
                            pixmap = QtGui.QPixmap(fname)
                            self.thumbs.setPixmap(pixmap)
                            self.thumbs.adjustSize()
                        self.thumbs.setFrameShape(QtGui.QFrame.StyledPanel)
                        return

        self.thumbs.setPixmap(QtGui.QPixmap())
        self.thumbs.setFrameShape(QtGui.QFrame.NoFrame)

###############################################################################

class QVersionMashups(QtGui.QWidget):
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.versionNumber = None
        self.controller = None
        #label = QtGui.QLabel("Mashups:")
        self.mashupsButton = QtGui.QToolButton()
        self.mashupsButton.setText("Mashups")
        self.mashupsButton.setPopupMode(QtGui.QToolButton.InstantPopup)
        self.mashupsButton.setArrowType(QtCore.Qt.RightArrow)
        self.mashupsButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        
#        font = QtGui.QFont("Arial", 11, QtGui.QFont.Normal)
#        font.setItalic(True)
#        self.group = QtGui.QGroupBox()
#        helplabel = QtGui.QLabel("Double-click a mashup to execute it")
#        helplabel.setFont(font)
#        self.mashupsList = QtGui.QListWidget()
#        self.mashupsList.setSelectionMode(QtGui.QAbstractItemView.SingleSelection)
#        self.mashupsList.itemDoubleClicked.connect(self.mashupDoubleClicked)
#        glayout = QtGui.QVBoxLayout()
#        glayout.setMargin(2)
#        glayout.setSpacing(5)
#        #layout.addWidget(label)
#        glayout.addWidget(self.mashupsList)
#        glayout.addWidget(helplabel)
#        self.group.setLayout(glayout)
        layout = QtGui.QVBoxLayout()
        layout.setMargin(2)
        layout.setSpacing(5)
        #layout.addStretch()
        layout.addWidget(self.mashupsButton)
        #layout.addWidget(self.group)
        self.setLayout(layout)
        self.apps = {}
        #self.group.setVisible(False)
        #self.connect(self.mashupsButton, QtCore.SIGNAL("clicked(bool)"),
        #             self.showList)
        
    def createMashupsMenu(self, tagMap):
        tags = tagMap.keys()
        self.mashupsButton.setText("Mashups (%s)"%str(len(tags)))
        #latestversion = mtrail.getLatestVersion()
        mashupsMenu = QtGui.QMenu(self)
        if len(tags) > 0:
            tags.sort()
            for tag in tags:
                action = QtGui.QAction(str(tag), self, 
                                       triggered=self.mashupSelected)
                action.setData(tagMap[tag])
                mashupsMenu.addAction(action)
            self.mashupsButton.setEnabled(True)
        return mashupsMenu
    
#    def showList(self, on):
#        self.group.setVisible(on)
        
    def updateController(self, controller):
        """ updateController(controller: VistrailController) -> None

        """
        self.controller = controller

    def updateVersion(self, versionNumber):
        """ updateVersion(versionNumber: int) -> None

        """
        #self.mashupsList.clear()
        
        from vistrails.gui.mashups.mashups_manager import MashupsManager
        from vistrails.gui.mashups.mashups_inspector import QMashupListPanelItem
        getMshptrail = MashupsManager.getMashuptrailforVersionInVistrailController
        if self.controller:
            vistrail = self.controller.vistrail
            if versionNumber in vistrail.actionMap.keys():
                self.mtrail = getMshptrail(self.controller, versionNumber)
                if self.mtrail:
                    tagMap = self.mtrail.getTagMap()
                    self.mashupsButton.setMenu(self.createMashupsMenu(tagMap))
#                    tags = tagMap.keys()
#                    self.mashupsButton.setText("Mashups (%s)"%str(len(tags)))
#                    #latestversion = mtrail.getLatestVersion()
#                    if len(tags) > 0:
#                        tags.sort()
#                        for tag in tags:
#                            item = QMashupListPanelItem(str(tag),
#                                                        tagMap[tag],
#                                                        self.mashupsList)
#                        self.mashupsButton.setEnabled(True)
#                    else:
#                        self.mashupsButton.setEnabled(False)
                else:
                    self.mashupsButton.setText("Mashups (0)")
                    self.mashupsButton.setEnabled(False)
            else:
                self.mashupsButton.setText("Mashups (0)")
                self.mashupsButton.setEnabled(False)
        else:
            self.mashupsButton.setText("Mashups (0)")
            self.mashupsButton.setEnabled(False)
                
    def mashupSelected(self):
        action = self.sender()
        version, ok = action.data()
        self.openMashup(version)

    def openMashup(self, version):
        from vistrails.gui.mashups.mashups_manager import MashupsManager
        item_key = (self.mtrail.id, version)
        if self.apps.has_key(item_key):
            app = self.apps[item_key]
            if app:
                app.activateWindow()
                return
        
        app = MashupsManager.createMashupApp(self.controller, self.mtrail, 
                                             version)
        app.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        app.appWasClosed.connect(self.appWasClosed)
        self.apps[item_key] = app
        app.activateWindow()
        app.raise_()
                
    def appWasClosed(self, app):
        for (k, a) in self.apps.iteritems():
            if a == app:
                self.apps[k] = None



# pragma: no testimport
name = 'test_import_pkg'
identifier = 'org.vistrails.tests.test_import_pkg'
version = '0.42'

import vistrails.tests.resources.import_targets.test1
import tests.resources.import_targets.test2

import module1



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file has the implementation of an algorithm to layout general
rooted trees in a nice way. The "lw" in the file name "tree_layout_lw.py"
stands for (L)inear (W)alker. This code is based on the paper:

    Christoph Buchheim, Michael Junger, and Sebastian Leipert.
    Improving walker's algorithm to run in linear time.
    In Stephen G. Kobourov and Michael T. Goodrich, editors, Graph
    Drawing, volume 2528 of Lecture Notes in Computer Science, pages
    344-353. Springer, 2002.

which is a faster (it is linear!) way to compute the tree layout
proposed by Walker than the algorithm described by him.
The original paper is:

    John Q. Walker II.
    A node-positioning algorithm for general trees.
    Softw., Pract. Exper., 20(7):685-705, 1990.

"""

from __future__ import division

class TreeLW(object):
    """
    The input to the algorithm must be a tree
    in this format.

    """
    def __init__(self):
        self.nodes = []
        self.maxLevel = 0

    def root(self):
        return self.nodes[0]

    def addNode(self, parentNode, width, height, object = None):
        newNode = NodeLW(width,height,object)        
        self.nodes.append(newNode)

        # add
        if parentNode is not None:
            parentNode.addChild(newNode)

        # update max level
        self.maxLevel = max(self.maxLevel,newNode.level)
        return newNode

    def changeParentOfNodeWithNoParent(self, parentNode, childNode):
        if childNode.parent is not None:
            raise ValueError("Node already has a parent")

        parentNode.addChild(childNode)
        maxLevel = self.__dfsUpdateLevel(childNode)

        # update max level
        self.maxLevel = max(self.maxLevel, maxLevel)

    def __dfsUpdateLevel(self, node):
        if node.parent is None:
            node.level = 0
        else:
            node.level = node.parent.level + 1
        maxLevel = node.level
        for child in node.children:
            maxLevel = max(maxLevel, self.__dfsUpdateLevel(child))
        return maxLevel

    def boundingBox(self):
        kbb = KeepBoundingBox()
        for w in self.nodes:
            kbb.addPoint(w.x-w.width/2.0, w.y-w.height/2.0)
            kbb.addPoint(w.x+w.width/2.0, w.y+w.height/2.0)
        return kbb.getBoundingBox()            

    def getMaxNodeHeightPerLevel(self):
        result = [0] * (self.maxLevel+1)
        for w in self.nodes:
            level = w.level
            result[level] = max(result[level],w.height)
        return result

    @staticmethod
    def randomTree(n,  k=10000000):
        p = [0] * n
        import random
        for i in xrange(1, n):
            minIndex = max(i-k, 0)
            index = random.randint(minIndex, i-1) # random number in {0,1,2,...,i-1}
            p[i] = index
        t = TreeLW()
        nodes= []
        for i in xrange(n):
            if i==0:
                parent= None
            else:
                parent=nodes[p[i]]
            width = 5 + 10*random.random()
            height = 5 + 10*random.random()
            nodes.append(t.addNode(parent,width,height,i))
        return t            

class KeepBoundingBox(object):
    def __init__(self):
        self.minx = None
        self.miny = None
        self.maxx = None
        self.maxy = None
        self.size = 0

    def addPoint(self,x,y):
        if self.minx is None or self.minx > x:
            self.minx = x
        if self.miny is None or self.miny > y:
            self.miny = y
        if self.maxx is None or self.maxx < x:
            self.maxx = x
        if self.maxy is None or self.maxy < y:
            self.maxy = y
        self.size = self.size + 1

    def getBoundingBox(self):
        return [self.minx, self.miny, self.maxx-self.minx, self.maxy - self.miny]

class NodeLW(object):
    """
    Node of the tree with all the auxiliar
    variables needed to the LW algorithm.
    The fields width, height and object
    are given as input. The first two are
    used to layout while the last one might
    be used by the user of this class.

    """    
    def __init__(self, width, height, object = None):
        self.width = width
        self.height = height
        self.object = object

        self.children = []

        self.parent = None
        self.index = 0

        # level of the node
        self.level = 0
        
        # intermediate variables for 
        # layout algorithm
        self.mod = 0
        self.prelim = 0
        self.ancestor = None
        self.thread = None
        self.change = 0
        self.shift = 0

        # final center position
        self.x = 0
        self.y = 0
        
    def getNumChilds(self):
        return len(self.children)
        
    def hasChild(self):
        return bool(self.children)

    def addChild(self, node):
        self.children.append(node)
        node.index = len(self.children) - 1
        node.parent = self
        node.level = self.level + 1

    def isLeaf(self):
        return not self.children

    def leftChild(self):
        return self.children[0]

    def rightChild(self):
        return self.children[-1]

    def leftSibling(self):
        if self.index > 0:
            return self.parent.children[self.index-1]
        else:
            return None

    def leftMostSibling(self):
        if self.parent is not None:
            return self.parent.children[0]
        else:
            return self

    def isSiblingOf(self, v):
        return self.parent == v.parent and self.parent is not None

class TreeLayoutLW(object):

    """
    TreeLayoutLW: the LW stands for Linear Walker.

    This code is based on the paper:

    Christoph Buchheim, Michael Junger, and Sebastian Leipert.
    Improving walker's algorithm to run in linear time.
    In Stephen G. Kobourov and Michael T. Goodrich, editors, Graph
    Drawing, volume 2528 of Lecture Notes in Computer Science, pages
    344-353. Springer, 2002.

    which is a faster (linear) way to compute the tree layout
    proposed by Walker than the algorithm described by him.
    The original paper is:

    John Q. Walker II.
    A node-positioning algorithm for general trees.
    Softw., Pract. Exper., 20(7):685-705, 1990.
    """

    # vertical alignment of nodes in the 
    # level band: TOP, MIDDLE, BOTTOM
    TOP = 0
    MIDDLE = 1
    BOTTOM = 2

    def __init__(self, tree, vertical_alignment=1, xdistance=10, ydistance=10):
        self.xdistance = xdistance
        self.ydistance = ydistance
        self.tree = tree
        self.vertical_alignment = vertical_alignment
        self.treeLayout()

    def treeLayout(self):
        for v in self.tree.nodes:
            v.mod = 0
            v.thread = None
            v.ancestor = v
            
        r = self.tree.root()
        self.firstWalk(r)
        self.secondWalk(r, -r.prelim)
        self.setVerticalPositions()

    def setVerticalPositions(self):

        # set y position
        maxNodeHeightPerLevel = self.tree.getMaxNodeHeightPerLevel()
        info_level = []
        position_level = 0
        for level in xrange(len(maxNodeHeightPerLevel)):
            height_level = maxNodeHeightPerLevel[level]
            info_level.append((position_level,height_level))
            position_level += self.ydistance + height_level
            
        #
        for w in self.tree.nodes:
            level = w.level
            position_level, height_level = info_level[level]
            if self.vertical_alignment == TreeLayoutLW.TOP:
                w.y = position_level + w.height/2.0
            elif self.vertical_alignment == TreeLayoutLW.MIDDLE:
                w.y = position_level + height_level/2.0
            else: # bottom
                w.y = position_level + height_level - w.height/2.0
                
        
    def gap(self, v1, v2):

        return self.xdistance + (v1.width + v2.width)/2.0        


    def firstWalk(self, v):
        
        if v.isLeaf():
            v.prelim = 0
            w = v.leftSibling()
            if w is not None:
                v.prelim = w.prelim + self.gap(w,v)

        else:
            
            defaultAncestor = v.leftChild()
            for w in v.children:
                self.firstWalk(w)
                defaultAncestor = self.apportion(w, defaultAncestor)
            self.executeShifts(v)
            
            midpoint = (v.leftChild().prelim + v.rightChild().prelim) / 2.0

            w = v.leftSibling()
            if w is not None:
                v.prelim = w.prelim + self.gap(w,v)
                v.mod = v.prelim - midpoint
            else:
                v.prelim = midpoint


    def apportion(self,  v,  defaultAncestor):

        """
        Apportion: to divide and assign proportionally.

        Suppose the the left siblings of "v" are all aligned. 
        Now align the subtree with root "v" (note: the correct
        alignment of the left siblings of "v" are encoded 
        in the auxiliar variables, the x and y 
        are not correct; only in the end the correct 
        values are assigned to x  and y). 
        By property (*) in Section 4 the gratest 
        distinct ancestor of a node "w" in the 
        subtrees at rooted at the left siblings of "v"
        is w.ancestor if this value is a left sibling
        of v otherwise it is "defaultAncestor".

        """
        w = v.leftSibling()
        if w is not None:
            # p stands for + or plus (right subtree)
            # m stands for - or minus (left subtree)
            # i stands for inside
            # o stands for outside
            # v stands for vertex
            # s stands for shift
            vip = vop = v
            vim = w
            vom = vip.leftMostSibling()
            sip = vip.mod
            sop = vop.mod
            sim = vim.mod
            som = vom.mod
            while self.nextRight(vim) is not None and self.nextLeft(vip) is not None:
                
                vim = self.nextRight(vim)
                vip = self.nextLeft(vip)
                vom = self.nextLeft(vom)
                vop = self.nextRight(vop)

                vop.ancestor = v
                
                shift = (vim.prelim + sim) - (vip.prelim + sip) + self.gap(vim,vip)
                
                if shift > 0:
                    self.moveSubtree(self.ancestor(vim,v,defaultAncestor),v,shift)
                    sip += shift
                    sop += shift

                sim += vim.mod
                sip += vip.mod
                som += vom.mod
                sop += vop.mod

            if self.nextRight(vim) is not None and self.nextRight(vop) is None:
                vop.thread = self.nextRight(vim)
                vop.mod += sim - sop

            if self.nextLeft(vip) is not None and self.nextLeft(vom) is None:
                vom.thread = self.nextLeft(vip)
                vom.mod += sip - som
                defaultAncestor = v
            
        return defaultAncestor

    def nextLeft(self, v):
        if v.hasChild():
            return v.leftChild()
        else:
            return v.thread

    def nextRight(self, v):
        if v.hasChild():
            return v.rightChild()
        else:
            return v.thread

    def moveSubtree(self, wm, wp, shift):
        subtrees = float(wp.index - wm.index)
        wp.change += -shift/subtrees
        wp.shift += shift
        wm.change += shift/subtrees
        wp.prelim += shift
        wp.mod += shift

    def executeShifts(self, v):
        shift = 0
        change = 0
        for i in xrange(v.getNumChilds()-1,-1,-1):
            w = v.children[i]
            w.prelim += shift
            w.mod += shift
            change += w.change
            shift += w.shift + change

    def ancestor(self, vim, v, defaultAncestor):
        if vim.ancestor.isSiblingOf(v):
            return vim.ancestor
        else:
            return defaultAncestor

    def secondWalk(self,  v, m):
        v.x = v.prelim + m
        for w in v.children:
            self.secondWalk(w, m + v.mod)

# graph
if __name__ == "__main__":

    t = TreeLW()

    a = t.addNode(None,1,1,"a")

    b = t.addNode(a,1,1,"b")
    c = t.addNode(a,1,1,"c")
    d = t.addNode(a,1,1,"d")

    e = t.addNode(b,1,1,"e")
    f = t.addNode(b,1,1,"f")
    g = t.addNode(d,1,1,"g")
    h = t.addNode(d,1,1,"h")

    i = t.addNode(f,1,1,"i")
    j = t.addNode(f,1,1,"j")
    k = t.addNode(h,1,1,"k")
    l = t.addNode(h,1,1,"l")
    m = t.addNode(h,1,1,"m")
    n = t.addNode(h,1,1,"n")
    o = t.addNode(h,1,1,"o")
    
    layout = TreeLayoutLW(t)
    



#!/usr/bin/env python

# This example shows how to combine data from both the imaging and
# graphics pipelines. The vtkMergeFilter is used to merge the data
# from each together.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Read in an image and compute a luminance value. The image is
# extracted as a set of polygons (vtkImageDataGeometryFilter). We then
# will warp the plane using the scalar (luminance) values.
reader = vtk.vtkBMPReader()
reader.SetFileName(VTK_DATA_ROOT + "/Data/masonry.bmp")
luminance = vtk.vtkImageLuminance()
luminance.SetInputConnection(reader.GetOutputPort())
geometry = vtk.vtkImageDataGeometryFilter()
geometry.SetInputConnection(luminance.GetOutputPort())
warp = vtk.vtkWarpScalar()
warp.SetInputConnection(geometry.GetOutputPort())
warp.SetScaleFactor(-0.1)

# Use vtkMergeFilter to combine the original image with the warped
# geometry.
merge = vtk.vtkMergeFilter()
merge.SetGeometry(warp.GetOutput())
merge.SetScalars(reader.GetOutput())
mapper = vtk.vtkDataSetMapper()
mapper.SetInputConnection(merge.GetOutputPort())
mapper.SetScalarRange(0, 255)
mapper.ImmediateModeRenderingOff()
actor = vtk.vtkActor()
actor.SetMapper(mapper)

# Create renderer stuff
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(actor)
ren.ResetCamera()
ren.GetActiveCamera().Azimuth(20)
ren.GetActiveCamera().Elevation(30)
ren.SetBackground(0.1, 0.2, 0.4)
ren.ResetCameraClippingRange()

renWin.SetSize(250, 250)

cam1 = ren.GetActiveCamera()
cam1.Zoom(1.4)

iren.Initialize()
renWin.Render()
iren.Start()



#!/usr/bin/env python

# This example demonstrates the use of glyphing. We also use a mask filter
# to select a subset of points to glyph.

import vtk
from vtk.util.misc import vtkGetDataRoot
VTK_DATA_ROOT = vtkGetDataRoot()

# Read a data file. This originally was a Cyberware laser digitizer scan 
# of Fran J.'s face. Surface normals are generated based on local geometry
# (i.e., the polygon normals surrounding eash point are averaged). We flip
# the normals because we want them to point out from Fran's face.
fran = vtk.vtkPolyDataReader()
fran.SetFileName(VTK_DATA_ROOT + "/Data/fran_cut.vtk")
normals = vtk.vtkPolyDataNormals()
normals.SetInputConnection(fran.GetOutputPort())
normals.FlipNormalsOn()
franMapper = vtk.vtkPolyDataMapper()
franMapper.SetInputConnection(normals.GetOutputPort())
franActor = vtk.vtkActor()
franActor.SetMapper(franMapper)
franActor.GetProperty().SetColor(1.0, 0.49, 0.25)

# We subsample the dataset because we want to glyph just a subset of
# the points. Otherwise the display is cluttered and cannot be easily
# read. The RandonModeOn and SetOnRatio combine to random select one out
# of every 10 points in the dataset.
ptMask = vtk.vtkMaskPoints()
ptMask.SetInputConnection(normals.GetOutputPort())
ptMask.SetOnRatio(10)
ptMask.RandomModeOn()

# In this case we are using a cone as a glyph. We transform the cone so
# its base is at 0,0,0. This is the point where glyph rotation occurs.
cone = vtk.vtkConeSource()
cone.SetResolution(6)
transform = vtk.vtkTransform()
transform.Translate(0.5, 0.0, 0.0)
transformF = vtk.vtkTransformPolyDataFilter()
transformF.SetInputConnection(cone.GetOutputPort())
transformF.SetTransform(transform)

# vtkGlyph3D takes two inputs: the input point set (SetInput) which can be
# any vtkDataSet; and the glyph (SetSource) which must be a vtkPolyData.
# We are interested in orienting the glyphs by the surface normals that
# we previosuly generated.
glyph = vtk.vtkGlyph3D()
glyph.SetInputConnection(ptMask.GetOutputPort())
glyph.SetSource(transformF.GetOutput())
glyph.SetVectorModeToUseNormal()
glyph.SetScaleModeToScaleByVector()
glyph.SetScaleFactor(0.004)
spikeMapper = vtk.vtkPolyDataMapper()
spikeMapper.SetInputConnection(glyph.GetOutputPort())
spikeActor = vtk.vtkActor()
spikeActor.SetMapper(spikeMapper)
spikeActor.GetProperty().SetColor(0.0, 0.79, 0.34)

# Create the RenderWindow, Renderer and both Actors
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)

# Add the actors to the renderer, set the background and size
ren.AddActor(franActor)
ren.AddActor(spikeActor)

renWin.SetSize(500, 500)
ren.SetBackground(0.1, 0.2, 0.4)

# Set a nice camera position.
ren.ResetCamera()
cam1 = ren.GetActiveCamera()
cam1.Zoom(1.4)
cam1.Azimuth(110)

iren.Initialize()
renWin.Render()
iren.Start()
